from __future__ import unicode_literals

import tox

from .common import base_discover
from .via_path import check_with_path


@tox.hookimpl
def tox_get_python_executable(envconfig):
    spec, path = base_discover(envconfig)
    if path is not None:
        return path
    # 3. check if the literal base python
    candidates = [envconfig.basepython]
    # 4. check if the un-versioned name is good
    if spec.name is not None and spec.name != envconfig.basepython:
        candidates.append(spec.name)
    return check_with_path(candidates, spec)



from __future__ import absolute_import, unicode_literals

import signal
import subprocess
import sys
from datetime import datetime

import pytest
from flaky import flaky
from pathlib2 import Path

from tox.constants import INFO
from tox.util.main import MAIN_FILE


@flaky(max_runs=3)
@pytest.mark.skipif(INFO.IS_PYPY, reason="TODO: process numbers work differently on pypy")
@pytest.mark.skipif(
    "sys.platform == 'win32'", reason="triggering SIGINT reliably on Windows is hard",
)
def test_parallel_interrupt(initproj, monkeypatch, capfd):
    monkeypatch.setenv(str("_TOX_SKIP_ENV_CREATION_TEST"), str("1"))
    monkeypatch.setenv(str("TOX_REPORTER_TIMESTAMP"), str("1"))
    start = datetime.now()
    initproj(
        "pkg123-0.7",
        filedefs={
            "tox.ini": """
                    [tox]
                    envlist = a, b

                    [testenv]
                    skip_install = True
                    commands = python -c "open('{{envname}}', 'w').write('done'); \
                    import time; time.sleep(100)"
                    whitelist_externals = {}

                """.format(
                sys.executable,
            ),
        },
    )
    process = subprocess.Popen(
        [sys.executable, MAIN_FILE, "-p", "all"],
        creationflags=(
            subprocess.CREATE_NEW_PROCESS_GROUP
            if sys.platform == "win32"
            else 0
            # needed for Windows signal send ability (CTRL+C)
        ),
    )
    try:
        import psutil

        current_process = psutil.Process(process.pid)
    except ImportError:
        current_process = None

    wait_for_env_startup(process)

    all_children = []
    if current_process is not None:
        all_children.append(current_process)
        all_children.extend(current_process.children(recursive=True))
        assert len(all_children) >= 1 + 2 + 2, all_children
    end = datetime.now() - start
    assert end
    process.send_signal(signal.CTRL_C_EVENT if sys.platform == "win32" else signal.SIGINT)
    process.wait()
    out, err = capfd.readouterr()
    output = "{}\n{}".format(out, err)
    assert "KeyboardInterrupt parallel - stopping children" in output, output
    assert "ERROR:   a: parallel child exit code " in output, output
    assert "ERROR:   b: parallel child exit code " in output, output
    for process in all_children:
        msg = "{}{}".format(output, "\n".join(repr(i) for i in all_children))
        assert not process.is_running(), msg


def wait_for_env_startup(process):
    """the environments will write files once they are up"""
    signal_files = [Path() / "a", Path() / "b"]
    found = False
    while True:
        if process.poll() is not None:
            break
        for signal_file in signal_files:
            if not signal_file.exists():
                break
        else:
            found = True
            break
    if not found or process.poll() is not None:
        missing = [f for f in signal_files if not f.exists()]
        out, _ = process.communicate()
        assert len(missing), out
        assert False, out



# FIXME this seems unnecessary
# TODO move fixtures here and only keep helper functions/classes in the plugin
# TODO _pytest_helpers might be a better name than _pytestplugin then?
# noinspection PyUnresolvedReferences
from tox._pytestplugin import *  # noqa



"""Implement https://www.python.org/dev/peps/pep-0514/ to discover interpreters - Windows only"""
from __future__ import unicode_literals

import os
import re

import six
from six.moves import winreg

from tox import reporter
from tox.interpreters.py_spec import PythonSpec


def enum_keys(key):
    at = 0
    while True:
        try:
            yield winreg.EnumKey(key, at)
        except OSError:
            break
        at += 1


def get_value(key, value_name):
    try:
        return winreg.QueryValueEx(key, value_name)[0]
    except OSError:
        return None


def discover_pythons():
    for hive, hive_name, key, flags, default_arch in [
        (winreg.HKEY_CURRENT_USER, "HKEY_CURRENT_USER", r"Software\Python", 0, 64),
        (
            winreg.HKEY_LOCAL_MACHINE,
            "HKEY_LOCAL_MACHINE",
            r"Software\Python",
            winreg.KEY_WOW64_64KEY,
            64,
        ),
        (
            winreg.HKEY_LOCAL_MACHINE,
            "HKEY_LOCAL_MACHINE",
            r"Software\Python",
            winreg.KEY_WOW64_32KEY,
            32,
        ),
    ]:
        for spec in process_set(hive, hive_name, key, flags, default_arch):
            yield spec


def process_set(hive, hive_name, key, flags, default_arch):
    try:
        with winreg.OpenKeyEx(hive, key, 0, winreg.KEY_READ | flags) as root_key:
            for company in enum_keys(root_key):
                if company == "PyLauncher":  # reserved
                    continue
                for spec in process_company(hive_name, company, root_key, default_arch):
                    yield spec
    except OSError:
        pass


def process_company(hive_name, company, root_key, default_arch):
    with winreg.OpenKeyEx(root_key, company) as company_key:
        for tag in enum_keys(company_key):
            for spec in process_tag(hive_name, company, company_key, tag, default_arch):
                yield spec


def process_tag(hive_name, company, company_key, tag, default_arch):
    with winreg.OpenKeyEx(company_key, tag) as tag_key:
        major, minor = load_version_data(hive_name, company, tag, tag_key)
        if major is None:
            return
        arch = load_arch_data(hive_name, company, tag, tag_key, default_arch)
    exe, args = load_exe(hive_name, company, company_key, tag)
    if exe is not None:
        name = "python" if company == "PythonCore" else company
        yield PythonSpec(name, major, minor, arch, exe, args)


def load_exe(hive_name, company, company_key, tag):
    key_path = "{}/{}/{}".format(hive_name, company, tag)
    try:
        with winreg.OpenKeyEx(company_key, r"{}\InstallPath".format(tag)) as ip_key:
            with ip_key:
                exe = get_value(ip_key, "ExecutablePath")
                if exe is None:
                    ip = get_value(ip_key, None)
                    if ip is None:
                        msg(key_path, "no ExecutablePath or default for it")

                    else:
                        exe = os.path.join(ip, "python.exe")
                if os.path.exists(exe):
                    args = get_value(ip_key, "ExecutableArguments")
                    return exe, args
                else:
                    msg(key_path, "exe does not exists {}".format(exe))
    except OSError:
        msg("{}/{}".format(key_path, "InstallPath"), "missing")
    return None, None


def load_arch_data(hive_name, company, tag, tag_key, default_arch):
    arch_str = get_value(tag_key, "SysArchitecture")
    if arch_str is not None:
        key_path = "{}/{}/{}/SysArchitecture".format(hive_name, company, tag)
        try:
            return parse_arch(arch_str)
        except ValueError as sys_arch:
            msg(key_path, sys_arch)
    return default_arch


def parse_arch(arch_str):
    if not isinstance(arch_str, six.string_types):
        raise ValueError("arch is not string")
    match = re.match(r"(\d+)bit", arch_str)
    if match:
        return int(next(iter(match.groups())))
    raise ValueError("invalid format {}".format(arch_str))


def load_version_data(hive_name, company, tag, tag_key):
    version_str = get_value(tag_key, "SysVersion")
    major, minor = None, None
    if version_str is not None:
        key_path = "{}/{}/{}/SysVersion".format(hive_name, company, tag)
        try:
            major, minor = parse_version(get_value(tag_key, "SysVersion"))
        except ValueError as sys_version:
            msg(key_path, sys_version)
    if major is None:
        key_path = "{}/{}/{}".format(hive_name, company, tag)
        try:
            major, minor = parse_version(tag)
        except ValueError as tag_version:
            msg(key_path, tag_version)
    return major, minor


def parse_version(version_str):
    if not isinstance(version_str, six.string_types):
        raise ValueError("key is not string")
    match = re.match(r"(\d+)\.(\d+).*", version_str)
    if match:
        return tuple(int(i) for i in match.groups())
    raise ValueError("invalid format {}".format(version_str))


def msg(path, what):
    reporter.verbosity1("PEP-514 violation in Windows Registry at {} error: {}".format(path, what))


def _run():
    reporter.update_default_reporter(0, reporter.Verbosity.DEBUG)
    for spec in discover_pythons():
        print(repr(spec))


if __name__ == "__main__":
    _run()



def test_parallel_inception(initproj, cmd):
    initproj(
        "inception-1.2.3",
        filedefs={
            # the outer config just has one env: graham
            "tox.ini": """
            [tox]
            envlist = graham
            skipsdist = True

            [testenv]
            commands =
                python runner.py
            """,
            # the inner config has 3 different envs, 1 of them is graham
            "inner": {
                "tox.ini": """
                [tox]
                envlist = graham,john,terry
                skipsdist = True

                [testenv]
                commands =
                    python -c 'pass'
                """,
            },
            # the outer test runs the inner tox and asserts all 3 envs were run
            "runner.py": """
            import os
            import subprocess
            import sys

            os.chdir("inner")
            p = subprocess.Popen(("tox"), stdout=subprocess.PIPE, universal_newlines=True)
            stdout, _ = p.communicate()
            sys.stdout.write(stdout)
            assert "graham" in stdout
            assert "john" in stdout
            assert "terry" in stdout
            """,
        },
        add_missing_setup_py=False,
    )

    result = cmd("-p", "all", "-o")
    result.assert_success()

    # 1 from the outer, 1 from the inner
    assert result.out.count("graham: commands succeeded") == 2
    # those gentlemen are only inside
    assert result.out.count("john: commands succeeded") == 1
    assert result.out.count("terry: commands succeeded") == 1



import sys

dist_folder = sys.argv[1]
backend_spec = sys.argv[2]
backend_obj = sys.argv[3] if len(sys.argv) >= 4 else None

backend = __import__(backend_spec, fromlist=[None])
if backend_obj:
    backend = getattr(backend, backend_obj)

basename = backend.build_sdist(dist_folder, {"--global-option": ["--formats=gztar"]})
print(basename)



import sys
import threading
from contextlib import contextmanager
from tempfile import TemporaryFile

if sys.version_info >= (3, 8):
    from importlib import metadata as importlib_metadata  # noqa
else:
    import importlib_metadata  # noqa


def is_main_thread():
    """returns true if we are within the main thread"""
    cur_thread = threading.current_thread()
    if sys.version_info >= (3, 4):
        return cur_thread is threading.main_thread()
    else:
        # noinspection PyUnresolvedReferences
        return isinstance(cur_thread, threading._MainThread)


# noinspection PyPep8Naming
@contextmanager
def suppress_output():
    """suppress both stdout and stderr outputs"""
    if sys.version_info >= (3, 5):
        from contextlib import redirect_stdout, redirect_stderr
    else:

        class _RedirectStream(object):

            _stream = None

            def __init__(self, new_target):
                self._new_target = new_target
                self._old_targets = []

            def __enter__(self):
                self._old_targets.append(getattr(sys, self._stream))
                setattr(sys, self._stream, self._new_target)
                return self._new_target

            def __exit__(self, exctype, excinst, exctb):
                setattr(sys, self._stream, self._old_targets.pop())

        class redirect_stdout(_RedirectStream):
            _stream = "stdout"

        class redirect_stderr(_RedirectStream):
            _stream = "stderr"

    with TemporaryFile("wt") as file:
        with redirect_stdout(file):
            with redirect_stderr(file):
                yield



import os

from tox.util import set_os_env_var


def test_set_os_env_var_clean_env(monkeypatch):
    monkeypatch.delenv("ENV", raising=False)
    with set_os_env_var("ENV", "a"):
        assert os.environ["ENV"] == "a"
    assert "ENV" not in os.environ


def test_set_os_env_var_exist_env(monkeypatch):
    monkeypatch.setenv("ENV", "b")
    with set_os_env_var("ENV", "a"):
        assert os.environ["ENV"] == "a"
    assert os.environ["ENV"] == "b"


def test_set_os_env_var_non_str():
    with set_os_env_var("ENV", 1):
        assert os.environ["ENV"] == "1"



# -*- coding: utf-8 -*-
"""A minimal non-colored version of https://pypi.org/project/halo, to track list progress"""
from __future__ import absolute_import, unicode_literals

import os
import sys
import threading
from collections import OrderedDict
from datetime import datetime

import py

threads = []

if os.name == "nt":
    import ctypes

    class _CursorInfo(ctypes.Structure):
        _fields_ = [("size", ctypes.c_int), ("visible", ctypes.c_byte)]


def _file_support_encoding(chars, file):
    encoding = getattr(file, "encoding", None)
    if encoding is not None:
        for char in chars:
            try:
                char.encode(encoding)
            except UnicodeEncodeError:
                break
        else:
            return True
    return False


class Spinner(object):
    CLEAR_LINE = "\033[K"
    max_width = 120
    UNICODE_FRAMES = ["⠋", "⠙", "⠹", "⠸", "⠼", "⠴", "⠦", "⠧", "⠇", "⠏"]
    ASCII_FRAMES = ["|", "-", "+", "x", "*"]

    def __init__(self, enabled=True, refresh_rate=0.1):
        self.refresh_rate = refresh_rate
        self.enabled = enabled
        self._file = sys.stdout
        self.frames = (
            self.UNICODE_FRAMES
            if _file_support_encoding(self.UNICODE_FRAMES, sys.stdout)
            else self.ASCII_FRAMES
        )
        self.stream = py.io.TerminalWriter(file=self._file)
        self._envs = OrderedDict()
        self._frame_index = 0

    def clear(self):
        if self.enabled:
            self.stream.write("\r")
            self.stream.write(self.CLEAR_LINE)

    def render(self):
        while True:
            self._stop_spinner.wait(self.refresh_rate)
            if self._stop_spinner.is_set():
                break
            self.render_frame()
        return self

    def render_frame(self):
        if self.enabled:
            self.clear()
            self.stream.write("\r{}".format(self.frame()))

    def frame(self):
        frame = self.frames[self._frame_index]
        self._frame_index += 1
        self._frame_index = self._frame_index % len(self.frames)
        text_frame = "[{}] {}".format(len(self._envs), " | ".join(self._envs))
        if len(text_frame) > self.max_width - 1:
            text_frame = "{}...".format(text_frame[: self.max_width - 1 - 3])
        return "{} {}".format(*[(frame, text_frame)][0])

    def __enter__(self):
        if self.enabled:
            self.disable_cursor()
        self.render_frame()
        self._stop_spinner = threading.Event()
        self._spinner_thread = threading.Thread(target=self.render)
        self._spinner_thread.setDaemon(True)
        self._spinner_thread.start()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if not self._stop_spinner.is_set():
            if self._spinner_thread:
                self._stop_spinner.set()
                self._spinner_thread.join()

            self._frame_index = 0
            if self.enabled:
                self.clear()
                self.enable_cursor()

        return self

    def add(self, name):
        self._envs[name] = datetime.now()

    def succeed(self, key):
        self.finalize(key, "✔ OK", green=True)

    def fail(self, key):
        self.finalize(key, "✖ FAIL", red=True)

    def skip(self, key):
        self.finalize(key, "⚠ SKIP", white=True)

    def finalize(self, key, status, **kwargs):
        start_at = self._envs[key]
        del self._envs[key]
        if self.enabled:
            self.clear()
        self.stream.write(
            "{} {} in {}{}".format(
                status, key, td_human_readable(datetime.now() - start_at), os.linesep,
            ),
            **kwargs
        )
        if not self._envs:
            self.__exit__(None, None, None)

    def disable_cursor(self):
        if self._file.isatty():
            if os.name == "nt":
                ci = _CursorInfo()
                handle = ctypes.windll.kernel32.GetStdHandle(-11)
                ctypes.windll.kernel32.GetConsoleCursorInfo(handle, ctypes.byref(ci))
                ci.visible = False
                ctypes.windll.kernel32.SetConsoleCursorInfo(handle, ctypes.byref(ci))
            elif os.name == "posix":
                self.stream.write("\033[?25l")

    def enable_cursor(self):
        if self._file.isatty():
            if os.name == "nt":
                ci = _CursorInfo()
                handle = ctypes.windll.kernel32.GetStdHandle(-11)
                ctypes.windll.kernel32.GetConsoleCursorInfo(handle, ctypes.byref(ci))
                ci.visible = True
                ctypes.windll.kernel32.SetConsoleCursorInfo(handle, ctypes.byref(ci))
            elif os.name == "posix":
                self.stream.write("\033[?25h")


def td_human_readable(delta):
    seconds = int(delta.total_seconds())
    periods = [
        ("year", 60 * 60 * 24 * 365),
        ("month", 60 * 60 * 24 * 30),
        ("day", 60 * 60 * 24),
        ("hour", 60 * 60),
        ("minute", 60),
        ("second", 1),
    ]

    texts = []
    for period_name, period_seconds in periods:
        if seconds > period_seconds or period_seconds == 1:
            period_value, seconds = divmod(seconds, period_seconds)
            if period_name == "second":
                ms = delta.total_seconds() - int(delta.total_seconds())
                period_value = round(period_value + ms, 3)
            has_s = "s" if period_value != 1 else ""
            texts.append("{} {}{}".format(period_value, period_name, has_s))
    return ", ".join(texts)



from __future__ import absolute_import, unicode_literals

import os
import shutil
import subprocess
import sys

import py
import pytest
from pathlib2 import Path
from six.moves.urllib.parse import urljoin
from six.moves.urllib.request import pathname2url

from tox.exception import BadRequirement, MissingRequirement


@pytest.fixture(scope="session")
def next_tox_major():
    """a tox version we can guarantee to not be available"""
    return "10.0.0"


def test_provision_min_version_is_requires(newconfig, next_tox_major):
    with pytest.raises(MissingRequirement) as context:
        newconfig(
            [],
            """\
            [tox]
            minversion = {}
            """.format(
                next_tox_major,
            ),
        )
    config = context.value.config

    deps = [r.name for r in config.envconfigs[config.provision_tox_env].deps]
    assert deps == ["tox >= {}".format(next_tox_major)]
    assert config.run_provision is True
    assert config.toxworkdir
    assert config.toxinipath
    assert config.provision_tox_env == ".tox"
    assert config.ignore_basepython_conflict is False


def test_provision_tox_change_name(newconfig):
    config = newconfig(
        [],
        """\
        [tox]
        provision_tox_env = magic
        """,
    )
    assert config.provision_tox_env == "magic"


def test_provision_basepython_global_only(newconfig, next_tox_major):
    """we don't want to inherit basepython from global"""
    with pytest.raises(MissingRequirement) as context:
        newconfig(
            [],
            """\
            [tox]
            minversion = {}
            [testenv]
            basepython = what
            """.format(
                next_tox_major,
            ),
        )
    config = context.value.config
    base_python = config.envconfigs[".tox"].basepython
    assert base_python == sys.executable


def test_provision_basepython_local(newconfig, next_tox_major):
    """however adhere to basepython when explicilty set"""
    with pytest.raises(MissingRequirement) as context:
        newconfig(
            [],
            """\
            [tox]
            minversion = {}
            [testenv:.tox]
            basepython = what
            """.format(
                next_tox_major,
            ),
        )
    config = context.value.config
    base_python = config.envconfigs[".tox"].basepython
    assert base_python == "what"


def test_provision_bad_requires(newconfig, capsys, monkeypatch):
    with pytest.raises(BadRequirement):
        newconfig(
            [],
            """\
            [tox]
            requires = sad >sds d ok
            """,
        )
    out, err = capsys.readouterr()
    assert "ERROR: failed to parse InvalidRequirement" in out
    assert not err


@pytest.fixture()
def plugin(monkeypatch, tmp_path):
    dest = tmp_path / "a"
    shutil.copytree(str(py.path.local(__file__).dirpath().join("plugin")), str(dest))
    subprocess.check_output([sys.executable, "setup.py", "egg_info"], cwd=str(dest))
    monkeypatch.setenv(str("PYTHONPATH"), str(dest))


def test_provision_cli_args_ignore(cmd, initproj, monkeypatch, plugin):
    import tox.config
    import tox.session

    prev_ensure = tox.config.ParseIni.ensure_requires_satisfied

    @staticmethod
    def ensure_requires_satisfied(config, requires, min_version):
        result = prev_ensure(config, requires, min_version)
        config.run_provision = True
        return result

    monkeypatch.setattr(
        tox.config.ParseIni, "ensure_requires_satisfied", ensure_requires_satisfied,
    )
    prev_get_venv = tox.session.Session.getvenv

    def getvenv(self, name):
        venv = prev_get_venv(self, name)
        venv.envconfig.envdir = py.path.local(sys.executable).dirpath().dirpath()
        venv.setupenv = lambda: True
        venv.finishvenv = lambda: True
        return venv

    monkeypatch.setattr(tox.session.Session, "getvenv", getvenv)
    initproj("test-0.1", {"tox.ini": "[tox]"})
    result = cmd("-a", "--option", "b")
    result.assert_success(is_run_test_env=False)


def test_provision_cli_args_not_ignored_if_provision_false(cmd, initproj):
    initproj("test-0.1", {"tox.ini": "[tox]"})
    result = cmd("-a", "--option", "b")
    result.assert_fail(is_run_test_env=False)


@pytest.fixture(scope="session")
def wheel(tmp_path_factory):
    """create a wheel for a project"""
    state = {"at": 0}

    def _wheel(path):
        state["at"] += 1
        dest_path = tmp_path_factory.mktemp("wheel-{}-".format(state["at"]))
        env = os.environ.copy()
        try:
            subprocess.check_output(
                [
                    sys.executable,
                    "-m",
                    "pip",
                    "wheel",
                    "-w",
                    str(dest_path),
                    "--no-deps",
                    str(path),
                ],
                universal_newlines=True,
                stderr=subprocess.STDOUT,
                env=env,
            )
        except subprocess.CalledProcessError as exception:
            assert not exception.returncode, exception.output

        wheels = list(dest_path.glob("*.whl"))
        assert len(wheels) == 1
        wheel = wheels[0]
        return wheel

    return _wheel


THIS_PROJECT_ROOT = Path(__file__).resolve().parents[3]


@pytest.fixture(scope="session")
def tox_wheel(wheel):
    return wheel(THIS_PROJECT_ROOT)


@pytest.fixture(scope="session")
def magic_non_canonical_wheel(wheel, tmp_path_factory):
    magic_proj = tmp_path_factory.mktemp("magic")
    (magic_proj / "setup.py").write_text(
        "from setuptools import setup\nsetup(name='com.magic.this-is-fun')",
    )
    return wheel(magic_proj)


def test_provision_non_canonical_dep(
    cmd, initproj, monkeypatch, tox_wheel, magic_non_canonical_wheel,
):
    initproj(
        "w-0.1",
        {
            "tox.ini": """\
            [tox]
            envlist = py
            requires =
                com.magic.this-is-fun
                tox == {}
            [testenv:.tox]
            passenv = *
            """.format(
                tox_wheel.name.split("-")[1],
            ),
        },
    )
    find_links = " ".join(
        space_path2url(d) for d in (tox_wheel.parent, magic_non_canonical_wheel.parent)
    )

    monkeypatch.setenv(str("PIP_FIND_LINKS"), str(find_links))

    result = cmd("-a", "-v", "-v")
    result.assert_success(is_run_test_env=False)


def test_provision_requirement_with_environment_marker(cmd, initproj):
    initproj(
        "proj",
        {
            "tox.ini": """\
            [tox]
            requires =
                package-that-does-not-exist;python_version=="1.0"
            """,
        },
    )
    result = cmd("-e", "py", "-vv")
    result.assert_success(is_run_test_env=False)


def space_path2url(path):
    at_path = str(path)
    if " " not in at_path:
        return at_path
    return urljoin("file:", pathname2url(os.path.abspath(at_path)))



import subprocess

import pytest


def need_executable(name, check_cmd):
    def wrapper(fn):
        try:
            subprocess.check_output(check_cmd)
        except OSError:
            return pytest.mark.skip(reason="{} is not available".format(name))(fn)
        return fn

    return wrapper


def need_git(fn):
    return pytest.mark.git(need_executable("git", ("git", "--version"))(fn))



import functools
import os
import signal
import socket
import sys

import py
import pytest

import tox
from tox.logs import ResultLog


@pytest.fixture(name="pkg")
def create_fake_pkg(tmpdir):
    pkg = tmpdir.join("hello-1.0.tar.gz")
    pkg.write("whatever")
    return pkg


@pytest.fixture()
def clean_hostname_envvar(monkeypatch):
    monkeypatch.delenv("HOSTNAME", raising=False)
    return functools.partial(monkeypatch.setenv, "HOSTNAME")


def test_pre_set_header(clean_hostname_envvar):
    replog = ResultLog()
    d = replog.dict
    assert replog.dict == d
    assert replog.dict["reportversion"] == "1"
    assert replog.dict["toxversion"] == tox.__version__
    assert replog.dict["platform"] == sys.platform
    assert replog.dict["host"] == socket.getfqdn()
    data = replog.dumps_json()
    replog2 = ResultLog.from_json(data)
    assert replog2.dict == replog.dict


def test_set_header(pkg, clean_hostname_envvar):
    replog = ResultLog()
    d = replog.dict
    assert replog.dict == d
    assert replog.dict["reportversion"] == "1"
    assert replog.dict["toxversion"] == tox.__version__
    assert replog.dict["platform"] == sys.platform
    assert replog.dict["host"] == socket.getfqdn()
    expected = {"basename": "hello-1.0.tar.gz", "sha256": pkg.computehash("sha256")}
    env_log = replog.get_envlog("a")
    env_log.set_header(installpkg=pkg)
    assert env_log.dict["installpkg"] == expected

    data = replog.dumps_json()
    replog2 = ResultLog.from_json(data)
    assert replog2.dict == replog.dict


def test_hosname_via_envvar(clean_hostname_envvar):
    clean_hostname_envvar("toxicity")
    replog = ResultLog()
    assert replog.dict["host"] == "toxicity"


def test_addenv_setpython(pkg):
    replog = ResultLog()
    envlog = replog.get_envlog("py36")
    envlog.set_python_info(py.path.local(sys.executable))
    envlog.set_header(installpkg=pkg)
    assert envlog.dict["python"]["version_info"] == list(sys.version_info)
    assert envlog.dict["python"]["version"] == sys.version
    assert envlog.dict["python"]["executable"] == sys.executable


def test_get_commandlog(pkg):
    replog = ResultLog()
    envlog = replog.get_envlog("py36")
    assert "setup" not in envlog.dict
    setuplog = envlog.get_commandlog("setup")
    envlog.set_header(installpkg=pkg)
    setuplog.add_command(["virtualenv", "..."], "venv created", 0)
    expected = [{"command": ["virtualenv", "..."], "output": "venv created", "retcode": 0}]
    assert setuplog.list == expected
    assert envlog.dict["setup"]
    setuplog2 = replog.get_envlog("py36").get_commandlog("setup")
    assert setuplog2.list == setuplog.list


@pytest.mark.parametrize("exit_code", [None, 0, 5, 128 + signal.SIGTERM, 1234, -15])
@pytest.mark.parametrize("os_name", ["posix", "nt"])
def test_invocation_error(exit_code, os_name, mocker, monkeypatch):
    monkeypatch.setattr(os, "name", value=os_name)
    mocker.spy(tox.exception, "exit_code_str")
    result = str(tox.exception.InvocationError("<command>", exit_code=exit_code))
    # check that mocker works, because it will be our only test in
    # test_z_cmdline.py::test_exit_code needs the mocker.spy above
    assert tox.exception.exit_code_str.call_count == 1
    call_args = tox.exception.exit_code_str.call_args
    assert call_args == mocker.call("InvocationError", "<command>", exit_code)
    if exit_code is None:
        assert "(exited with code" not in result
    elif exit_code == -15:
        assert "(exited with code -15 (SIGTERM))" in result
    else:
        assert "(exited with code {})".format(exit_code) in result
        note = "Note: this might indicate a fatal error signal"
        if (os_name == "posix") and (exit_code == 128 + signal.SIGTERM):
            assert note in result
            assert "({} - 128 = {}: SIGTERM)".format(exit_code, signal.SIGTERM) in result
        else:
            assert note not in result



from __future__ import unicode_literals

import json
import os
import subprocess
from collections import defaultdict
from threading import Lock

import py

from tox import reporter
from tox.constants import VERSION_QUERY_SCRIPT

from .py_spec import PythonSpec


def check_with_path(candidates, spec):
    for path in candidates:
        base = path
        if not os.path.isabs(path):
            path = py.path.local.sysfind(path)
        if path is not None:
            if os.path.exists(str(path)):
                cur_spec = exe_spec(path, base)
                if cur_spec is not None and cur_spec.satisfies(spec):
                    return cur_spec.path


_SPECS = {}
_SPECK_LOCK = defaultdict(Lock)


def exe_spec(python_exe, base):
    if not isinstance(python_exe, str):
        python_exe = str(python_exe)
    with _SPECK_LOCK[python_exe]:
        if python_exe not in _SPECS:
            info = get_python_info(python_exe)
            if info is not None:
                found = PythonSpec(
                    "pypy" if info["implementation"] == "PyPy" else "python",
                    info["version_info"][0],
                    info["version_info"][1],
                    64 if info["is_64"] else 32,
                    info["executable"],
                )
                reporter.verbosity2("{} ({}) is {}".format(base, python_exe, info))
            else:
                found = None
            _SPECS[python_exe] = found
    return _SPECS[python_exe]


_python_info_cache = {}


def get_python_info(cmd):
    try:
        return _python_info_cache[cmd].copy()
    except KeyError:
        pass
    proc = subprocess.Popen(
        [cmd] + [VERSION_QUERY_SCRIPT],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        universal_newlines=True,
    )
    out, err = proc.communicate()
    if not proc.returncode:
        try:
            result = json.loads(out)
        except ValueError as exception:
            failure = exception
        else:
            _python_info_cache[cmd] = result
            return result.copy()
    else:
        failure = "exit code {}".format(proc.returncode)
    reporter.verbosity1("{!r} cmd {!r} out {!r} err {!r} ".format(failure, cmd, out, err))
